/**************************************************************************************

      watsup.cpp    Linux system and process monitoring program
      Free software licensed under GNU General Public License v.2
      source URL: http://kornelix.squarespace.com/watsup

***************************************************************************************/

#define  gtitle "watsup v.09"

#include <dirent.h>
#include <sys/resource.h>
#include "zfuncs.h"

//  Linux system parameters
int      pagesize;                                                         //  memory page size
double   jiffy;                                                            //  CPU time slice
int      ncpus;                                                            //  CPU count

//  program parameters and limits
#define  maxpid 1000                                                       //  max. PIDs sampled
#define  maxtop 100                                                        //  max. top PIDs to report
#define  maxdev 8                                                          //  max. CPUs / disks to report
#define  proc_ignore "getty init"                                          //  processes to ignore
#define  mega (1024 * 1024)                                                //  computer megabyte

//  editable parameters and default values
double   interval = 1.0;                                                   //  sample interval, 1 sec.
int      fontsize = 8;                                                     //  report font size
char     headerfont[20] = "Monospace bold 8";                              //  report header font
char     bodyfont[20] = "Monospace 8";                                     //  report body font

//  calculated parameters                                                  //  v.03
int      winX = 600, winY = 400;                                           //  report window dimensions
int      fontheight = 20;                                                  //  report font height
int      topmost = 20;                                                     //  topmost procs within window

GtkWidget      *mWin, *mVbox, *mScroll, *mLog;                             //  main window widgets

struct   sysrec {                                                          //  system statistics record layout
   double   cpubusy[maxdev];
   double   memused, memfree, memcache;
   double   pfsoft, pfhard;
   double   diskops, disksecs;
   double   netrecv, netxmit;
};

sysrec      srec1, srec2;                                                  //  system stats, sample 1 and sample 2

struct   pidrec {                                                          //  process statistics record layout
   int      pid;
   int      row;
   double   weight;
   double   cpu;
   double   mem;
   double   pfs;
   double   dmbs;
   char     username[20];
   char     filename[40];
};

pidrec      prec1[maxpid];                                                 //  proc recs, sample 1
pidrec      prec2[maxpid];                                                 //  proc recs, sample 2
pidrec      prec3[maxpid];                                                 //  proc recs for report
int         nrec1, nrec2, nrec3;                                           //  record counts for precN arrays
int         rec1, rec2, rec3;                                              //  indexes for precN arrays
int         repline;                                                       //  tracks report output lines

pthread_t   reptid = 0;                                                    //  report thread ID
timeval     time1, time2;                                                  //  sample times 1 and 2
double      elapsed;                                                       //  elapsed time between 1 and 2
int         intervalchanged = 0;                                           //  flag, interval changed by user
int         resortflag = 0;                                                //  flag, re-sort topmost by weight

int initfunc(void * data);                                                 //  GTK initial function
int m_expose();                                                            //  main window expose event

void menufunc(GtkWidget *item, const char *menu);                          //  menu functions
void m_parms();
void m_start();
void m_stop();
void m_sort();
void m_save();
void m_print();
void m_quit();
void m_help();

void get_sys_stats(sysrec *srec);                                          //  functions for system statistics
void report_sys_stats();
void copy_sys_stats();

void get_proc_stats(pidrec *prec, int &nnp);                               //  functions for process statistics
void report_proc_stats();
void copy_proc_stats();

int parseProcFile(cchar *pfile, cchar *pname, double *value, ...);         //  parse /proc/xxxx data
int parseProcRecord(char *prec, int field, double *value, ...);

void setreportfonts();


//  main windowing program

int main(int argc, char *argv[])
{
   GtkWidget   *tbar;

   gtk_init(&argc,&argv);                                                  //  initz. GTK
   if (! g_thread_supported())                                             //  suddenly required for new gtk   v.05
         g_thread_init(0);                                                 //  initz. GTK for threads
   gdk_threads_init();   
   zlockInit();                                                            //  initz. for threads

   initz_appfiles("watsup",null);                                          //  set up app directories

   mWin = gtk_window_new(GTK_WINDOW_TOPLEVEL);                             //  main window
   gtk_window_set_title(GTK_WINDOW(mWin),gtitle);
   gtk_window_set_position(GTK_WINDOW(mWin),GTK_WIN_POS_CENTER);
   gtk_window_set_default_size(GTK_WINDOW(mWin),winX,winY);
   
   mVbox = gtk_vbox_new(0,0);                                              //  vertical packing box
   gtk_container_add(GTK_CONTAINER(mWin),mVbox);                           //  add to main window

   tbar = create_toolbar(mVbox);                                           //  create tool bar and buttons
   add_toolbar_button(tbar,"parms","edit parameters","gtk-preferences",menufunc);
   add_toolbar_button(tbar,"start","start reporting","gtk-execute",menufunc);
   add_toolbar_button(tbar,"stop","stop reporting","gtk-stop",menufunc);
   add_toolbar_button(tbar,"sort","resort top processes","gtk-refresh",menufunc);
   add_toolbar_button(tbar,"save","save report to a file","gtk-save",menufunc);
   add_toolbar_button(tbar,"print","print report","gtk-print",menufunc);
   add_toolbar_button(tbar,"quit","quit watsup","gtk-quit",menufunc);
   add_toolbar_button(tbar,"help","view user guide","gtk-help",menufunc);

   mScroll = gtk_scrolled_window_new(0,0);                                 //  scrolled window
   gtk_box_pack_start(GTK_BOX(mVbox),mScroll,1,1,0);                       //  add to main window mVbox
   mLog = gtk_text_view_new();                                             //  text window
   gtk_container_add(GTK_CONTAINER(mScroll),mLog);                         //  add to scrolled window

   G_SIGNAL(mWin,"destroy",m_quit,0)                                       //  connect window destroy event
   G_SIGNAL(mWin,"expose-event",m_expose,0)

   gtk_init_add((GtkFunction) initfunc,0);                                 //  setup initial call from gtk_main()

   gtk_widget_show_all(mWin);                                              //  show all widgets

   gdk_threads_enter();
   gtk_main();                                                             //  process GUI events
   gdk_threads_leave();

   return 0;
}


//  initial function called from gtk_main() at startup

int initfunc(void * data)
{
   int      err, prio;

   jiffy = 1.0 / sysconf(_SC_CLK_TCK);                                     //  "jiffy" time slice = 1.0 / HZ
   pagesize = sysconf(_SC_PAGESIZE);                                       //  system page size
   ncpus = sysconf(_SC_NPROCESSORS_ONLN);                                  //  SMP CPU count
   if (ncpus > maxdev) zappcrash("too many CPUs: %d",ncpus);

   prio = getpriority(PRIO_PROCESS,0);                                     //  current priority
   if (prio > -2 && getuid() == 0) {
      err = setpriority(PRIO_PROCESS,0,prio-2);                            //  boost priority if allowed
      if (err) printf("watsup: unable to boost priority \n");
   }

   setreportfonts();                                                       //  setup report fonts from font size
   m_start();                                                              //  start report running
   return 0;
}


//  window expose event - recalibrate report window dimensions   v.03

int m_expose()
{
   GdkWindow   *textwin;

   textwin = gtk_text_view_get_window(GTK_TEXT_VIEW(mLog),GTK_TEXT_WINDOW_WIDGET);
   gdk_drawable_get_size(textwin,&winX,&winY);
   return 0;
}


//  menu selection events

void menufunc(GtkWidget *item, const char *menu)
{
   if (strEqu(menu,"parms")) m_parms();
   if (strEqu(menu,"start")) m_start();
   if (strEqu(menu,"stop")) m_stop();
   if (strEqu(menu,"sort")) m_sort();
   if (strEqu(menu,"save")) m_save();
   if (strEqu(menu,"print")) m_print();
   if (strEqu(menu,"quit")) m_quit();
   if (strEqu(menu,"help")) m_help();
   return;
}


//  edit report parameters

void m_parms() 
{ 
   zdialog     *zd;
   
   zd = zdialog_new("edit watsup parameters","done","cancel",null);
   zdialog_add_widget(zd,"hbox","hb1","dialog");
   zdialog_add_widget(zd,"vbox","vb1","hb1",0,"homog | space=5");          //  v.04
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"homog | space=5");
   zdialog_add_widget(zd,"label","lab1","vb1","sample period");
   zdialog_add_widget(zd,"label","lab3","vb1","font size");
   zdialog_add_widget(zd,"spin","period","vb2","0.1|99.0|0.1|1.0");
   zdialog_add_widget(zd,"spin","font","vb2","5|20|1|9");
   
   zdialog_stuff(zd,"period",interval);                                    //  stuff widgets with current params
   zdialog_stuff(zd,"font",fontsize);

   int zd_parms_compl(zdialog *zd, int zstat);                             //  dialog completion function
   zdialog_run(zd, 0, zd_parms_compl);                                     //  run dialog (non-modal)
   return;
}

int zd_parms_compl(zdialog *zd, int zstat)                                 //  dialog completion function
{
   double      interval2;

   if (zstat == 1) {
      zdialog_fetch(zd,"period",interval2);                                //  get new params from widgets
      zdialog_fetch(zd,"font",fontsize);
      
      if (interval2 != interval) {                                         //  if new sample interval,
         intervalchanged++;                                                //    notify reort thread
         interval = interval2;
      }
      
      setreportfonts();                                                    //  calculate font metrics
   }

   zdialog_destroy(zd);
   return 1;
}


//  set up report fonts from user font size setting, recalculate font metrics    v.03

void setreportfonts()
{
   PangoLanguage           *plang;
   PangoFontDescription    *pfontdesc;
   PangoContext            *pcontext;
   PangoFont               *pfont;
   PangoFontMetrics        *pmetrics;
   int                     fontascent, fontdescent;
   
   snprintf(headerfont,19,"Monospace bold %d",fontsize);                   //  report header font, bold
   snprintf(bodyfont,19,"Monospace %d",fontsize);                          //  report body font

   pfontdesc = pango_font_description_from_string(bodyfont);               //  set default font for report window
   gtk_widget_modify_font(mLog,pfontdesc);

   plang = pango_language_from_string("en-US");                            //  get font metrics (what a mess)
   pcontext = gtk_widget_get_pango_context(mLog);
   pfont = pango_context_load_font(pcontext,pfontdesc);
   pmetrics = pango_font_get_metrics(pfont,plang);
   fontascent = pango_font_metrics_get_ascent(pmetrics) / PANGO_SCALE;
   fontdescent = pango_font_metrics_get_descent(pmetrics) / PANGO_SCALE;
   fontheight = fontascent + fontdescent;                                  //  effective report line height
   return;
}

   
//  start reporting

void m_start()
{
   void * report_thread(void *);
   int      err;
   
   if (reptid) return;
   err = pthread_create(&reptid,0,report_thread,0);                        //  launch report thread
   if (err) zappcrash("thread create failure");
   return;
}


//  stop reporting

void m_stop()
{
   if (reptid) pthread_cancel(reptid);                                     //  cancel report thread
   reptid = 0;
   return;
}


//  re-sort top processes to top of list

void m_sort()
{ 
   resortflag = 1;                                                         //  set flag for process report
   return;
}


//  exit immediately

void m_quit()
{
   m_stop();                                                               //  stop thread
   gtk_main_quit();                                                        //  exit program
   return;
}


//  report thread function

void * report_thread(void *)
{
   double   interval2 = interval;                                          //  adjusted interval

   pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED,0);

   gettimeofday(&time1,0);                                                 //  get time #1
   get_sys_stats(&srec1);                                                  //  get system stats #1
   get_proc_stats(prec1,nrec1);                                            //  get process stats #1
   
   while (true)
   {
      zsleep(interval2);                                                   //  wait for interval
      pthread_testcancel();

      elapsed = get_timer(&time1);                                         //  precise elapsed time
      interval2 = interval2 - 0.5 * (elapsed - interval);                  //  adjust toward target interval

      if (intervalchanged) {                                               //  interval changed by user
         interval2 = interval;
         intervalchanged = 0;
      }

      gettimeofday(&time2,0);                                              //  get time #2
      get_sys_stats(&srec2);                                               //  get system stats #2
      get_proc_stats(prec2,nrec2);                                         //  get process stats #2

      report_sys_stats();                                                  //  system report for interval #1 to #2
      report_proc_stats();                                                 //  process report for interval #1 to #2

      time1 = time2;                                                       //  copy time #2 to time #1
      copy_sys_stats();                                                    //  copy system stats #2 to #1
      copy_proc_stats();                                                   //  copy process stats #2 to #1
   }
}


//  get system level statistics from various /proc/xxxx files

void get_sys_stats(sysrec *srec)
{
   int         ii, count;
   FILE        *fid;
   char        *pp, buff[1000];
   double      val1, val2, val3, val4;
   
   memset(srec,0,sizeof(sysrec));                                          //  zero data
   
   fid = fopen("/proc/stat","r");                                          //  read /proc/stat
   if (fid) {
      while ((pp = fgets(buff,999,fid)))
      {
         if (strncmp(pp,"cpu",3) == 0) {                                   //  look for "cpuN" record
            ii = atoi(pp+3);
            if (ii < 0 || ii >= ncpus) continue;
            parseProcRecord(pp,2,&val1,4,&val2,null);                      //  get user and system mode jiffies
            srec->cpubusy[ii] = val1 + val2;
            if (ii == ncpus-1) break;
         }
      }
      fclose(fid);
   }

   parseProcFile("/proc/meminfo","MemTotal:",&val1,"MemFree:",&val2,       //  memory statistics
                                 "Cached:",&val3,null);
   srec->memused = val1 - val2;
   srec->memfree = val2;
   srec->memcache = val3;

   parseProcFile("/proc/vmstat","pgfault",&val1,"pgmajfault",&val2,null);  //  page fault statistics
   srec->pfsoft = val1;
   srec->pfhard = val2;

   fid = fopen("/proc/diskstats","r");                                     //  read disk statistics
   if (fid) {
      while ((pp = fgets(buff,999,fid)))
      {
         pp = strstr(buff,"sda");                                          //  look for disks: "sda" "hda"
         if (! pp) pp = strstr(buff,"hda");
         if (! pp) continue;
         if (pp[3] != ' ') continue;                                       //  ignore partitions: "sdaN" "hdaN"

         count = parseProcRecord(pp,2,&val1,4,&val2,6,&val3,8,&val4,null); //  get read/write counts and sectors
         srec->diskops += val1 + val3;
         srec->disksecs += val2 + val4;
      }
      fclose(fid);
   }
   
   fid = fopen("/proc/net/dev","r");                                       //  read network statistics
   if (fid) {
      while ((pp = fgets(buff,999,fid)))
      {
         pp = strchr(buff,':');                                            //  look for device "eth0:" etc.
         if (! pp) continue;
         count = parseProcRecord(pp+1,1,&val1,9,&val2,null);               //  get bytes received and transmitted
         srec->netrecv += val1;
         srec->netxmit += val2;
      }
      fclose(fid);
   }
   
   return;   
}


//  read all /proc/<pid>/stat files and extract statistics data
      
void get_proc_stats(pidrec *prec, int &nnp)
{
   DIR            *procdir;
   FILE           *fid;
   struct dirent  *procent;
   char           *ppid, *pp, username[20], filename[40];
   char           buff1[40], buff2[1000];
   int            ii, cc, pid;
   double         cpu1, cpu2, mem, pfs, rbs, wbs;

   procdir = opendir("/proc");
   if (! procdir) zappcrash("open /proc failed: %s",syserrText());
   
   nnp = 0;

   while (true)                                                            //  loop per PID
   {
      procent = readdir(procdir);                                          //  read /proc directory
      if (! procent) break;
      ppid = procent->d_name;
      if (*ppid < '1' || *ppid > '9') continue;                            //  skip non-pid entries
      
      pid = atoi(ppid);
      pfs = cpu1 = cpu2 = mem = 0;
      *filename = *username = 0;

      snprintf(buff1,39,"/proc/%s/environ",ppid);                          //  read file /proc/<pid>/environ
      fid = fopen(buff1,"r");
      if (! fid) continue;
      pp = fgets(buff2,999,fid);
      fclose(fid);
      if (! pp) continue;

      strcpy(username,"root");
      for (ii = 0; ii < 10; ii++) {                                        //  kludge:                   v.07
         if (! *pp) break;                                                 //    root processes have no
         if (strnEqu(pp,"USER=",5)) {                                      //    USER=name env. variable
            strncpy0(username,pp+5,19);                                    //    and UID not in /proc/<pid>
            break; }
         pp += strlen(pp) + 1;
      }

      snprintf(buff1,39,"/proc/%s/exe",ppid);                              //  proc/<pid>/exe
      cc = readlink(buff1,buff2,999);                                      //  executable pathname/filename
      if (cc <= 0) continue;
      buff2[cc] = 0;
      pp = strrchr(buff2,'/');                                             //  isolate filename
      if (pp) pp++;
      else pp = buff2;
      if (strstr(proc_ignore,pp)) continue;                                //  ignore some processes
      strncpy0(filename,pp,39);                                            //  keep filename only

      snprintf(buff1,31,"/proc/%s/stat",ppid);                             //  read file /proc/<pid>/stat
      fid = fopen(buff1,"r");
      if (! fid) continue;
      pp = fgets(buff2,999,fid);
      fclose(fid);
      if (! pp) continue;
      
      pp = strchr(pp,')');                                                 //  closing ')' after (short) filename
      if (pp) 
         parseProcRecord(pp+1,10,&pfs,12,&cpu1,13,&cpu2,22,&mem,null);     //  parse and get wanted fields
      
      snprintf(buff1,39,"/proc/%s/io",ppid);
      parseProcFile(buff1,"read-bytes",&rbs,"write-bytes",&wbs,null);      //  get physical I/O byte counts

      prec[nnp].pid = pid;                                                 //  build proc record
      prec[nnp].cpu = cpu1 + cpu2;                                         //  user and system mode jiffies
      prec[nnp].mem = mem;                                                 //  real memory
      prec[nnp].pfs = pfs;                                                 //  major page faults
      prec[nnp].dmbs = (rbs + wbs);                                        //  I/O bytes 
      strcpy(prec[nnp].username,username);                                 //  user name              v.07
      strcpy(prec[nnp].filename,filename);                                 //  executable file
      
      if (++nnp == maxpid) appcrash("exceeded %d processes",maxpid);
   }

   closedir(procdir);
   return;
}


//  create report from two sets of system data and time interval

void report_sys_stats()
{
   int         ii, jj, lines;
   double      cpu1, cpupct, MBused, MBfree, MBcache;
   double      pfsoft, pfhard;
   double      diskops, diskMBs;
   double      netrecv, netxmit;
   char        sysline[maxdev*100], *pp;
   const char  *header = " period    CPU    memory MB     PFs/sec      disk/sec    net KB/sec \n";

   repline = 1;
   wprintx(mLog,repline++,"\n");
   wprintx(mLog,repline++,header,headerfont);

   for (cpupct = ii = 0; ii < ncpus; ii++) {                               //  accumulate total CPU load   v.08
      cpu1 = (srec2.cpubusy[ii] - srec1.cpubusy[ii]);                      //    from all SMP processors
      cpupct += cpu1 * 100.0 * jiffy / elapsed;
   }

   MBused = srec2.memused / 1024.0;                                        //  get used, free, cached memory, MB
   MBfree = srec2.memfree / 1024.0;
   MBcache = srec2.memcache / 1024.0;                                      //  v.03
   
   pfsoft = (srec2.pfsoft - srec1.pfsoft) / elapsed;                       //  hard and soft faults / sec
   pfhard = (srec2.pfhard - srec1.pfhard) / elapsed;
   
   diskops = (srec2.diskops - srec1.diskops) / elapsed;                    //  disk IOs and KB / sec
   diskMBs = (srec2.disksecs - srec1.disksecs) / elapsed;
   diskMBs = diskMBs / 2048;                                               //  512 byte sectors to megabytes  v.03

   netrecv = (srec2.netrecv - srec1.netrecv) / elapsed / 1024;             //  network KB / sec
   netxmit = (srec2.netxmit - srec1.netxmit) / elapsed / 1024;

   memset(sysline,' ',sizeof(sysline));                                    //  clear report lines

   lines = ncpus;                                                          //  report lines, enough room
   if (lines < 3) lines = 3;

   for (ii = 0; ii < lines; ii++)
   {
      pp = sysline + 100 * ii;                                             //  pointer to line 0, 1, ...
      
      if (ii == 0) sprintf(pp,"%7.2f",elapsed);                            //  elapsed time on line 0
      
      if (ii == 0) sprintf(pp+10,"%4.0f",cpupct);                          //  CPU % on line 0        v.08

      if (ii == 0) sprintf(pp+18,"used: %.0f",MBused);                     //  memory statistics on lines 0-2
      if (ii == 1) sprintf(pp+18,"free: %.0f",MBfree);
      if (ii == 2) sprintf(pp+17,"cache: %.0f",MBcache);
      
      if (ii == 0) sprintf(pp+32,"soft: %.0f",pfsoft);                     //  page fault stats on lines 0-1
      if (ii == 1) sprintf(pp+32,"hard: %.0f",pfhard);
      
      if (ii == 0) sprintf(pp+45,"IOs: %.0f",diskops);                     //  disk I/O stats on lines 0-1
      if (ii == 1) sprintf(pp+45,"MBs: %.0f",diskMBs);
      
      if (ii == 0) sprintf(pp+57,"recv: %.0f",netrecv);                    //  network stats on lines 0-1
      if (ii == 1) sprintf(pp+57,"xmit: %.0f",netxmit);
   }
   
   for (ii = 0; ii < lines; ii++)                                          //  output the lines
   {
      pp = sysline + 100 * ii;
      for (jj = 0; jj < 99; jj++) if (pp[jj] < ' ') pp[jj] = ' ';          //  remove imbedded nulls
      for (jj = 98; jj; jj--) if (pp[jj] > ' ') break;                     //  and trailing blanks
      pp[jj+1] = 0;
      wprintf(mLog,repline++,"%s \n",pp);
   }
}


//  create report from two sets of process data and time interval

void report_proc_stats()
{
   const char  *header1 = "                  real    hard    I/O   \n";
   const char  *header2 = "   PID     CPU     MBs    PF/s    MB/s   User     Executable \n";

   double      cpupct, MB, weight;
   char        scpupct[8];
   int         row, nrep;
   int         rowmap[maxpid];
                                                                           //  sort keys:
   int      key1[2][3] = { { 8, 8, 8 }, { 0, 4, 3 } };                     //    weight descending, PID ascending
   int      key2[1][3] = { { 4, 4, 3 } };                                  //    row ascending
   int      key3[1][3] = { { 0, 4, 3 } };                                  //    PID ascending

   wprintx(mLog,repline++,"\n");                                           //  output headers
   wprintx(mLog,repline++,header1,headerfont);
   wprintx(mLog,repline++,header2,headerfont);

   for (rec2 = 0; rec2 < nrec2; rec2++)                                    //  copy proc recs #2 to #3 for reporting
      prec3[rec2] = prec2[rec2];
   nrec3 = nrec2;

   for (rec1 = rec3 = 0; rec3 < nrec3; rec1++, rec3++)                     //  loop through proc recs #3
   {
      if (rec1 >= nrec1 || prec1[rec1].pid != prec3[rec3].pid) {           //  find same PID in proc recs #1
         for (rec1 = 0; rec1 < nrec1; rec1++)                              //  (utilize order, but don't count on it)
            if (prec1[rec1].pid == prec3[rec3].pid) break;
      }
      
      if (rec1 < nrec1) {                                                  //  if found, calculate deltas:
         prec3[rec3].cpu -= prec1[rec1].cpu;                               //    cpu jiffies
         prec3[rec3].pfs -= prec1[rec1].pfs;                               //    hard page faults
         prec3[rec3].dmbs -= prec1[rec1].dmbs;                             //    I/O bytes
         prec3[rec3].dmbs = prec3[rec3].dmbs / mega;                       //    I/O megabytes        v.03

         prec3[rec3].weight = prec1[rec1].weight;                          //  copy weight and row from proc rec #1
         prec3[rec3].row = prec1[rec1].row;                                //    to proc rec #3
      }

      else {
         prec3[rec3].weight = 0;                                           //  not found (new process was started)
         prec3[rec3].row = 0;                                              //  weight and row = 0
      }
   }

   for (rec3 = 0; rec3 < nrec3; rec3++)                                    //  calc. new sort weights
   {
      weight = prec3[rec3].cpu + 0.1 * prec3[rec3].pfs + prec3[rec3].dmbs;
      if (prec3[rec3].row) weight += 0.1;
      weight = 0.25 * weight + 0.75 * prec3[rec3].weight;                  //  use historical weighted average
      prec3[rec3].weight = weight;
   }
      
   MemSort((char *) prec3, sizeof(pidrec), nrec3, key1, 2);                //  sort proc recs by weight descending

   topmost = winY / fontheight - repline;                                  //  proc recs fitting in report window  v.03
   if (topmost < 3) topmost = 3;
   if (topmost < nrec3) nrep = topmost;                                    //  topmost proc recs to report
   else nrep = nrec3;

   for (row = 0; row < nrep; row++) rowmap[row] = 0;                       //  clear row map

   for (rec3 = 0; rec3 < nrep; rec3++)                                     //  assign report row to topmost proc recs
   {
      row = prec3[rec3].row;                                               //  keep prior report row
      if (row > 0 && row < nrep) rowmap[row-1] = 1;                        //    if available and in range
      else prec3[rec3].row = 0;
   }
   
   for (rec3 = row = 0; rec3 < nrep; rec3++)                               //  assign rest from top down
   {
      if (prec3[rec3].row > 0) continue;
      while (rowmap[row] > 0) row++;
      prec3[rec3].row = 1 + row++;
   }

   for (rec3 = nrep; rec3 < nrec3; rec3++)                                 //  clear all rows after topmost
      prec3[rec3].row = 0;
   
   if (resortflag) {                                                       //  if a re-sort was requested - 
      for (rec3 = 0; rec3 < nrep; rec3++)                                  //    keep current order by weight
            prec3[rec3].row = rec3 + 1;
      resortflag = 0;
   }

   MemSort((char *) prec3, sizeof(pidrec), nrep, key2, 1);                 //  sort topmost proc recs by row

   for (rec3 = 0; rec3 < nrep; rec3++)                                     //  output each proc record
   {
      cpupct = prec3[rec3].cpu * 100.0 * jiffy / elapsed;                  //  format 0.0 to 9.9, then 10 to 100
      if (cpupct < 9.95) sprintf(scpupct,"%4.1f",cpupct);
      else sprintf(scpupct,"%4.0f",cpupct);

      MB = prec3[rec3].mem * pagesize / mega;

      wprintf(mLog,repline++,"%6d  %6s  %6.1f  %6.0f  %6.1f   %-8s %s \n", 
               prec3[rec3].pid, scpupct, MB, prec3[rec3].pfs, 
               prec3[rec3].dmbs, prec3[rec3].username, prec3[rec3].filename);
   }

   wclear(mLog,repline);                                                   //  clear rest of window
   
   MemSort((char *) prec2, sizeof(pidrec), nrec2, key3, 1);                //  sort proc recs #2 by PID ascending
   MemSort((char *) prec3, sizeof(pidrec), nrec3, key3, 1);                //  sort proc recs #3 by PID ascending
   
   for (rec2 = 0; rec2 < nrec2; rec2++)                                    //  copy new weights and rows to proc recs #2
   {
      if (prec2[rec2].pid != prec3[rec2].pid) zappcrash("bug 23");
      prec2[rec2].weight = prec3[rec2].weight;
      prec2[rec2].row = prec3[rec2].row;
   }
   
   return;
}


//  copy system stats data set #2 to data set #1

void copy_sys_stats()
{
   srec1 = srec2;
}


//  copy proc stats data set #2 to data set #1

void copy_proc_stats()
{
   for (rec2 = 0; rec2 < nrec2; rec2++)
         prec1[rec2] = prec2[rec2];
   nrec1 = nrec2;
   return;
}


//  save report to a file      

void m_save()
{ 
   wfilesave(mLog);
   return;
}


//  print report on default printer

void m_print()
{ 
   wprintp(mLog);
   return;
}


//  thread function to display help file

void m_help()                                                              //  menu function
{
   showz_helpfile();                                                       //  launch help file in new process   v.02
   return;
}


//  Parse /proc record of the type  "xxx xxxxx xxxxx xxxxxxxx xxx"
//  Return numeric values for requested fields (starting with 1)

int parseProcRecord(char *prec, int field, double *value, ...)             //  EOL = 0
{
   va_list     arglist;
   int         xfield = 1, found = 0;

   va_start(arglist,value);

   while (*prec == ' ') prec++;                                            //  skip leading blanks

   while (field > 0)
   {
      while (xfield < field)
      {
         prec = strchr(prec,' ');                                          //  find next blank
         if (! prec) break;
         while (*prec == ' ') prec++;                                      //  skip multiple blanks
         xfield++;
      }
      
      if (! prec) break;
      *value = atof(prec);                                                 //  convert, return double 
      found++;

      field = va_arg(arglist,int);                                         //  next field number
      value = va_arg(arglist,double *);                                    //  next output double *
   }
   
   while (field > 0)
   {
      *value = 0;                                                          //  zero values not found
      field = va_arg(arglist,int);
      value = va_arg(arglist,double *);
   }
   
   va_end(arglist);
   return found;
}


//  Read and parse /proc file with records formatted "parmname xxxxxxx"
//  Find all requested parameters and return their numeric values

int parseProcFile(cchar *pfile, cchar *pname, double *value, ...)          //  EOL = 0
{
   FILE        *fid;
   va_list     arglist;
   char        buff[1000];
   const char  *pnames[20];
   double      *values[20];
   int         ii, fcc, wanted, found;
   
   pnames[0] = pname;                                                      //  1st parameter
   values[0] = value;
   *value = 0;
   
   va_start(arglist,value);

   for (ii = 1; ii < 20; ii++)                                             //  get all parameters
   {
      pnames[ii] = va_arg(arglist,char *);
      if (! pnames[ii]) break;
      values[ii] = va_arg(arglist,double *);
      *values[ii] = 0;                                                     //  initialize to zero
   }
      
   va_end(arglist);

   if (ii == 20) zappcrash("parseProcFile, too many fields");
   wanted = ii;
   found = 0;

   fid = fopen(pfile,"r");                                                 //  open /proc/xxx file
   if (! fid) return 0;

   while ((fgets(buff,999,fid)))                                           //  read record, "parmname nnnnn"
   {
      for (ii = 0; ii < wanted; ii++)
      {                                                                    //  look for my fields
         fcc = strlen(pnames[ii]);
         if (strnEqu(buff,pnames[ii],fcc)) {
            *values[ii] = atof(buff+fcc);                                  //  return value
            found++;
            break;
         }
      }

      if (found == wanted) break;                                          //  stop when all found
   }

   fclose(fid);
   return found;
}



